class SnakeTurret : ScriptObject
{
	float Radius = 7.f;
	float CooldownDuration = 1.5f;
	float ImpulseForce = 45.f;
	Node@ SnakeRagdollHolder = null;
	String TargetTag = "Enemy";
	String SnakeTag = "Weapon";
	
	// Is a weak handle since garbage collection is disabled
	// And to make sure our pointer becomes null
	protected WeakHandle TargetNode = null; // Node@
	protected Node@ LauncherNode = null;
	protected Node@ HotPoint = null;
	protected Node@ SnakeNode = null;
	// No PODVector, only Array
	protected Array<Drawable@> DrawableCache;
	protected bool IsLockedOn = false;
	protected bool IsLoaded = false;
	protected float StepAccumulator = 0.f;
	
	void DelayedStart() {
		SubscribeToEvent(node, "NodeCollision", "HandleNodeCollision");
		LauncherNode = node.GetChild("TurretHead");
		HotPoint = LauncherNode.GetChild("HotPoint");
		SetRadius(Radius);
	}

	void Update(float Timestep) {
		// If we're not locked don't update
		if (not IsLockedOn) return;
		// If target node is lost or has lost it's tag
		if (TargetNode.Get() is null or not cast<Node>(TargetNode.Get()).HasTag(TargetTag)) {
			// Nullify target in case
			TargetNode = null;
			IsLockedOn = false;
			return;
		}
		Vector3 Position = node.worldPosition;
		Vector3 Target = cast<Node>(TargetNode.Get()).worldPosition;

		// If target leaves range cleanup
		if (Position.DistanceToPoint(Target) > Radius) {
			TargetNode = null;
			IsLockedOn = false;
			Print("Target is too far");
			return;
		} else {
		//Print(Target);
		}
		Quaternion Rot;
		Rot.FromLookRotation(Target - Position);
		// Keep y rotation only
		Rot.x = 0;
		Rot.z = 0;
		LauncherNode.worldRotation = Rot;

		// If turret waited the cooldown time launch
		if (StepAccumulator >= CooldownDuration) {
			Reload();
			Launch();
			StepAccumulator = 0.f;
		} else {
			StepAccumulator += Timestep;
		}
	}

	void Launch() {
		if (not IsLoaded) Reload();
		// Reload failed
		if (not IsLockedOn or SnakeNode is null) return;

		// enable bodies
		Array<Component@> results;
		results = HotPoint.GetComponents("RigidBody", true);
		if (results.length <= 0) return;

		for (uint i = 0; i < results.length; i++) {
			cast<RigidBody>(results[i]).useGravity = true;
			cast<RigidBody>(results[i]).kinematic = false;
		}

		// Unparent (probably to a snake ragdoll holder)
		if(SnakeRagdollHolder is null) {
			SnakeNode.parent = node.scene;
		} else {
			SnakeNode.parent = SnakeRagdollHolder;
		}
		
		SnakeNode.AddTag(SnakeTag);

		// Apply force
		Vector3 Force = LauncherNode.worldRotation * Vector3::FORWARD * ImpulseForce;
		RigidBody@ body = SnakeNode.GetChild("Bone.001", true).GetComponent("RigidBody");
		body.ApplyImpulse(Force);
		// Get rid of node
		SnakeNode = null;
		IsLoaded = false;

		//log.Info("Shooting " + cast<Node>(TargetNode.Get()).name);
	}

	void Reload() {
		if (IsLoaded or (HotPoint is null)) return;
		// auto cache = GetSubsystem<ResourceCache>();

		SnakeNode = HotPoint.CreateChild("SnakeNode");
		SnakeNode.LoadXML(cast<XMLFile>(cache.GetResource("XMLFile", "SnakeTurret/Objects/Snek2.xml")).root);
		SnakeNode.position = Vector3::ZERO;
		//TODO: make rigid bodies disabled by default
		Array<Component@> results;
		results = HotPoint.GetComponents("RigidBody", true);
		if (results.length <= 0) return;

		for (uint i = 0; i < results.length; i++) {
			cast<RigidBody>(results[i]).useGravity = false;
			cast<RigidBody>(results[i]).kinematic = true;
		}
		IsLoaded = true;
	}

	void SetRadius(float range) {
		CollisionShape@ Collider = cast<CollisionShape>(node.GetComponent("CollisionShape"));
		Collider.size = Vector3(range, 0.1, 1);
		Radius = range;
	}

	void HandleNodeCollision(StringHash EventType, VariantMap& EventData) {
		if(TargetNode.Get() !is null) return;
		RigidBody@ OtherBody = EventData["OtherBody"].GetPtr();
		Node@ OtherNode = EventData["OtherNode"].GetPtr();
		if ((OtherBody is null) or (OtherNode is null)) return;
		if (not OtherNode.HasTag(TargetTag)) return;

		if (IsLockedOn) {
			return;
			// Check if is nearer
		} else {
			// If not then set lock on first target found
			TargetNode = OtherNode;
			IsLockedOn = true;
			//log.Info("Found new target: " + cast<Node>(TargetNode.Get()).name);
		}
	}
}